#ifndef __HISTORY_H_INCLUDED__
#define __HISTORY_H_INCLUDED__

int nlline_history_push(const char *line);
int nlline_history_pop(void);
int nlline_history_clear_all(void);
int nlline_history_current_init(void);
int nlline_history_current_next(void);
int nlline_history_current_prev(void);
char *nlline_history_get_current(void);
int nlline_history_is_current_start(void);
int nlline_history_is_current_end(void);

#endif
